/*
 *
 * Copyright (C) 2009 Red Hat, Inc.
 * Copyright (C) 2013 Collabora Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Written by: Arun Raghavan
 *
 */

#include "config.h"

#include <stdlib.h>
#include <pthread.h>

#include <gst/gst.h>

#include "ply-logger.h"
#include "ply-gst.h"

struct _ply_gst
{
  GstElement *pipeline;

  GMainContext *context;
  GMainLoop *mainloop;

  pthread_t thread;
  pthread_mutex_t mainloop_lock;
  pthread_cond_t mainloop_cond;
};

static void
mainloop_lock (ply_gst_t *g)
{
  pthread_mutex_lock (&g->mainloop_lock);
}

static void
mainloop_unlock (ply_gst_t *g)
{
  pthread_mutex_unlock (&g->mainloop_lock);
}

static void
mainloop_wait (ply_gst_t *g)
{
  pthread_cond_wait (&g->mainloop_cond, &g->mainloop_lock);
}

static void
mainloop_signal (ply_gst_t *g)
{
  mainloop_lock (g);
  pthread_cond_signal (&g->mainloop_cond);
  mainloop_unlock (g);
}


static gboolean
bus_watch (GstBus *bus, GstMessage *message, gpointer ptr)
{
  ply_gst_t *g = (ply_gst_t *) ptr;

  switch (GST_MESSAGE_TYPE (message))
    {
      case GST_MESSAGE_ASYNC_DONE:
      case GST_MESSAGE_ERROR:
	mainloop_signal (g);
	break;

      default:
	break;
    }

  return TRUE;
}

static void *
mainloop_thread (void *ptr)
{
  ply_gst_t *g = (ply_gst_t *) ptr;

  /* Wake up initialisation function now that the thread is running */
  mainloop_signal (g);

  g_main_context_push_thread_default (g->context);
  g_main_loop_run (g->mainloop);
  g_main_context_pop_thread_default (g->context);

  return 0;
}

ply_gst_t *
ply_gst_new (const char *pipeline)
{
  ply_gst_t *g = NULL;
  GError *error = NULL;
  GstBus *bus;

  ply_trace ("initialising gstreamer");

  if (!gst_init_check(NULL, NULL, &error))
    {
      ply_trace("Could not init GStreamer: %s", error->message);
      goto fail;
    }

  g = calloc (1, sizeof (ply_gst_t));

  ply_trace ("creating pipeline");

  g->pipeline = gst_parse_launch (pipeline, &error);
  if (error)
    {
      ply_trace ("Pipeline creation error: %s", error->message);
      goto fail;
    }
  if (!g->pipeline)
    goto fail;

  ply_trace ("created pipeline");

  /* Create the mainloop */
  g->context = g_main_context_new ();
  g->mainloop = g_main_loop_new (g->context, FALSE);


  /* Set up a bus watch */
  g_main_context_push_thread_default (g->context);
  bus = gst_pipeline_get_bus (GST_PIPELINE (g->pipeline));
  gst_bus_add_watch (bus, bus_watch, g);
  g_main_context_pop_thread_default (g->context);

  pthread_mutex_init (&g->mainloop_lock, NULL);
  pthread_cond_init (&g->mainloop_cond, NULL);

  /* Lock before starting the mainloop thread */
  mainloop_lock (g);

  if (pthread_create (&g->thread, NULL, mainloop_thread, g) < 0)
    {
      ply_trace ("Error creating thread");

      mainloop_unlock (g);

      pthread_mutex_destroy (&g->mainloop_lock);
      pthread_cond_destroy (&g->mainloop_cond);

      goto fail;
    }

  /* Wait till we know the thread's started */
  mainloop_wait (g);
  mainloop_unlock (g);

  return g;

fail:
  if (g)
    free (g);

  return NULL;
}

void
ply_gst_free (ply_gst_t *g)
{
  ply_gst_stop (g);

  gst_object_unref (g->pipeline);

  g_main_loop_quit (g->mainloop);
  g_main_loop_unref (g->mainloop);
  g_main_context_unref (g->context);

  pthread_join (g->thread, NULL);

  pthread_mutex_destroy (&g->mainloop_lock);
  pthread_cond_destroy (&g->mainloop_cond);

  free (g);
}

bool
ply_gst_start (ply_gst_t *g)
{
  GstStateChangeReturn ret;
  bool r = true;

  mainloop_lock (g);

  ret = gst_element_set_state (g->pipeline, GST_STATE_PLAYING);
  if (ret == GST_STATE_CHANGE_FAILURE)
    {
      ply_trace("Could not set pipeline to PLAYING");
      r = false;
    }
  else if (ret == GST_STATE_CHANGE_ASYNC)
    {
      /* Wait for state change to complete */
      mainloop_wait (g);
    }
  /* Success */
  mainloop_unlock (g);

  return r;
}

bool
ply_gst_stop (ply_gst_t *g)
{
  if (gst_element_set_state (g->pipeline, GST_STATE_NULL) ==
      GST_STATE_CHANGE_FAILURE)
    {
      ply_trace("Could not set pipeline to NULL");
    }
  return true;
}
